from dataclasses import dataclass

from terrex.net.streamer import Reader, Writer

from .net_module import NetServerModule


@dataclass()
class NetCreativePowerPermissionsModule(NetServerModule):
    id: int = 9
    zero: int = 0
    power_id: int = 0
    level: int = 0

    @classmethod
    def create(cls, zero: int, power_id: int, level: int) -> "NetCreativePowerPermissionsModule":
        obj = cls()
        obj.zero = zero
        obj.power_id = power_id
        obj.level = level
        return obj

    def read(self, reader: Reader) -> None:
        self.zero = reader.read_byte()
        if self.zero == 0:
            self.power_id = reader.read_ushort()
            self.level = reader.read_byte()

    def write(self, writer: Writer) -> None:
        writer.write_byte(self.zero)
        writer.write_ushort(self.power_id)
        writer.write_byte(self.level)



from dataclasses import dataclass

from terrex.net.streamer import Reader, Writer
from terrex.net.structure.bestiary import Bestiary

from .net_module import NetServerModule


@dataclass()
class NetBestiaryModule(NetServerModule):
    id: int = 4
    bestiary: Bestiary | None = None

    @classmethod
    def create(cls, bestiary: Bestiary) -> "NetBestiaryModule":
        obj = cls()
        obj.bestiary = bestiary
        return obj

    def read(self, reader: Reader) -> None:
        self.bestiary = Bestiary.read(reader)

    def write(self, writer: Writer) -> None:
        if self.bestiary is None:
            raise ValueError("bestiary must not be None")
        self.bestiary.write(writer)



"""
Ported from Terraria source: Terraria.WorldBuilding.Shapes.HalfCircle
"""

from math import sqrt

import numpy as np

from .base import GenAction, GenActionBulk, GenShape, Point


class HalfCircle(GenShape):
    """Half circle shape: upper half (default) or bottom half."""

    def __init__(self, radius: int, bottom_half: bool = False, quit_on_fail: bool = False):
        super().__init__(quit_on_fail)
        self._radius = radius
        self._bottomHalf = bottom_half

    def perform(self, origin: Point, action: GenAction) -> bool:
        num = (self._radius + 1) ** 2

        # Fast bulk path
        if isinstance(action, GenActionBulk):
            xs_all = []
            ys_all = []

            if self._bottomHalf:
                y_start, y_end = origin.y, origin.y + self._radius
            else:
                y_start, y_end = origin.y - self._radius, origin.y

            for i in range(y_start, y_end + 1):
                v = num - (i - origin.y) ** 2
                if v < 0:
                    continue
                num2 = min(self._radius, int(sqrt(v)))
                xs = np.arange(origin.x - num2, origin.x + num2 + 1)
                ys = np.full_like(xs, i)
                xs_all.append(xs)
                ys_all.append(ys)

            if not xs_all:
                return True

            return action.apply_bulk(np.concatenate(xs_all), np.concatenate(ys_all))

        # Standard path
        if self._bottomHalf:
            y_start, y_end = origin.y, origin.y + self._radius
        else:
            y_start, y_end = origin.y - self._radius, origin.y

        for i in range(y_start, y_end + 1):
            v = num - (i - origin.y) ** 2
            if v < 0:
                continue
            num2 = min(self._radius, int(sqrt(v)))
            for j in range(origin.x - num2, origin.x + num2 + 1):
                if not self.unit_apply(action, origin, j, i) and self._quitOnFail:
                    return False

        return True



"""
Ported from Terraria source: Terraria.WorldBuilding.GenShape and related base classes.

Original C# code adapted to Python with numpy optimizations for bulk operations.
"""

from dataclasses import dataclass

import numpy as np


@dataclass
class Point:
    """Equivalent to Microsoft.Xna.Framework.Point"""

    x: int
    y: int


@dataclass
class RectangleArea:
    """Equivalent to Microsoft.Xna.Framework.Rectangle (using left/top instead of X/Y)"""

    left: int = 0
    top: int = 0
    width: int = 0
    height: int = 0

    @property
    def right(self) -> int:
        return self.left + self.width

    @property
    def bottom(self) -> int:
        return self.top + self.height


class GenAction:
    """Base action applied to tile (x,y). Returns False on error."""

    def apply(self, x: int, y: int) -> bool:
        """
        Apply action to a single point.
        Returns False if operation failed.
        """
        raise NotImplementedError("Override apply()")


class GenShape:
    """Base shape. Perform generates points for action."""

    def __init__(self, quit_on_fail: bool = False):
        self._quitOnFail = quit_on_fail

    def unit_apply(self, action: GenAction, origin: Point, x: int, y: int) -> bool:
        """Internal UnitApply from C#."""
        return action.apply(x, y)

    def perform(self, origin: Point, action: GenAction) -> bool:
        """
        Execute shape: iterate over all points and apply action.
        Returns False if quit_on_fail and error.
        """
        raise NotImplementedError("Override Perform()")


class GenActionBulk(GenAction):
    """Base class for vectorized actions (numpy apply_bulk)."""

    def apply_bulk(self, xs: np.ndarray, ys: np.ndarray) -> bool:
        """
        Bulk application.
        Fallback: loop self.apply.
        """
        raise NotImplementedError("Override in subclass")


# Example usage:
# class BulkSetTile(GenActionBulk):
#     def __init__(self, world: np.ndarray, value):
#         self.world = world
#         self.value = value
#
#     def apply_bulk(self, xs, ys):
#         self.world[ys, xs] = self.value
#         return True



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.net.structure.vec2 import Vec2
from terrex.packet.base import SyncPacket


class SyncExtraValue(SyncPacket):
    id = MessageID.SyncExtraValue

    def __init__(self, npc_index: int = 0, extra_value: int = 0, pos: Vec2 | None = None):
        self.npc_index = npc_index
        self.extra_value = extra_value
        self.pos = pos or Vec2(0.0, 0.0)

    def read(self, reader: Reader):
        self.npc_index = reader.read_ushort()
        self.extra_value = reader.read_int()
        self.pos = Vec2.read(reader)

    def write(self, writer: Writer):
        writer.write_ushort(self.npc_index)
        writer.write_int(self.extra_value)
        self.pos.write(writer)



"""
PlayerTextureID - autogenerated from decompiled Terraria .cs files
Terraria namespace: Terraria.ID
Terraria version: v1.4.5.5
Generator: TerrariaPyGen v1.1.1 (Maksim Stashkevich)
Generated on: 2026-02-15 09:24
"""

from enum import IntEnum, auto


class PlayerTextureID(IntEnum):
    Head = 0
    EyeWhites = 1
    Eyes = 2
    TorsoSkin = 3
    Undershirt = 4
    Hands = 5
    Shirt = 6
    ArmSkin = 7
    ArmUndershirt = 8
    ArmHand = 9
    LegSkin = 10
    Pants = 11
    Shoes = 12
    ArmShirt = 13
    Extra = 14
    EyeBlink = 15




from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class UniqueTownNPCInfoSyncRequest(SyncPacket):
    id = MessageID.UniqueTownNPCInfoSyncRequest

    def __init__(
        self, npc_id: int = 0, name: str | None = None, town_npc_variation_idx: int | None = None
    ):
        self.npc_id = npc_id
        self.name = name
        self.town_npc_variation_idx = town_npc_variation_idx

    def read(self, reader: Reader) -> None:
        self.npc_id = reader.read_short()
        if not reader.eof():
            self.name = reader.read_dotnet_string()
            self.town_npc_variation_idx = reader.read_int()
        else:
            self.name = None
            self.town_npc_variation_idx = None

    def write(self, writer: Writer) -> None:
        writer.write_short(self.npc_id)
        if self.name is not None and self.town_npc_variation_idx is not None:
            writer.write_dotnet_string(self.name)
            writer.write_int(self.town_npc_variation_idx)



from terrex.id import MessageID
from terrex.packet.base import ClientPacket


class ClientSyncedInventory(ClientPacket):
    id = MessageID.ClientSyncedInventory

    def write(self, writer):
        pass



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ServerPacket


class ItemTweaker(ServerPacket):
    id = MessageID.ItemTweaker

    def __init__(self):
        self.item_index: int = 0
        self.flags1: int = 0
        self.packed_color: int = 0
        self.damage: int = 0
        self.knockback: float = 0.0
        self.use_animation: int = 0
        self.use_time: int = 0
        self.shoot: int = 0
        self.shoot_speed: float = 0.0
        self.flags2: int = 0
        self.width: int = 0
        self.height: int = 0
        self.scale: float = 0.0
        self.ammo: int = 0
        self.use_ammo: int = 0
        self.not_ammo: bool = False

    def read(self, reader: Reader) -> None:
        self.item_index = reader.read_short()
        self.flags1 = reader.read_byte()
        if self.flags1 & 1:
            self.packed_color = reader.read_int()
        if self.flags1 & 2:
            self.damage = reader.read_ushort()
        if self.flags1 & 4:
            self.knockback = reader.read_float()
        if self.flags1 & 8:
            self.use_animation = reader.read_ushort()
        if self.flags1 & 16:
            self.use_time = reader.read_ushort()
        if self.flags1 & 32:
            self.shoot = reader.read_short()
        if self.flags1 & 64:
            self.shoot_speed = reader.read_float()
        if self.flags1 & 128:
            self.flags2 = reader.read_byte()
            if self.flags2 & 1:
                self.width = reader.read_short()
            if self.flags2 & 2:
                self.height = reader.read_short()
            if self.flags2 & 4:
                self.scale = reader.read_float()
            if self.flags2 & 8:
                self.ammo = reader.read_short()
            if self.flags2 & 16:
                self.use_ammo = reader.read_short()
            if self.flags2 & 32:
                self.not_ammo = reader.read_bool()

    def write(self, writer: Writer) -> None:
        raise NotImplementedError("Server does not send TweakItem (client-bound packet only)")



from enum import IntEnum


class SkyEntityType(IntEnum):
    BirdsV = 0
    Wyvern = 1
    Airship = 2
    AirBalloon = 3
    Eyeball = 4
    Meteor = 5
    BoneSerpent = 6
    Bats = 7
    Butterflies = 8
    LostKite = 9
    Vulture = 10
    PixiePosse = 11
    Seagulls = 12
    SlimeBalloons = 13
    Gastropods = 14
    Pegasus = 15
    EaterOfSouls = 16
    Crimera = 17
    Hellbats = 18



from dataclasses import dataclass
from enum import IntEnum

from terrex.net.streamer import Reader, Writer

from .net_module import NetServerModule


class LeashedEntityMessageType(IntEnum):
    Remove = 0
    FullSync = 1
    PartialSync = 2


@dataclass()
class NetLeashedEntityModule(NetServerModule):
    id: int = 13
    msg_type: LeashedEntityMessageType | None = None
    slot: int | None = None
    leashed_type: int | None = None
    anchor: tuple[int, int] | None = None
    extra_data: bytes | None = None

    @classmethod
    def create(
        cls,
        msg_type: LeashedEntityMessageType,
        slot: int,
        leashed_type: int | None = None,
        anchor: tuple[int, int] | None = None,
        extra_data: bytes | None = None,
    ) -> "NetLeashedEntityModule":
        obj = cls()
        obj.msg_type = msg_type
        obj.slot = slot
        obj.leashed_type = leashed_type
        obj.anchor = anchor
        obj.extra_data = extra_data
        return obj

    def read(self, reader: Reader) -> None:
        self.msg_type = LeashedEntityMessageType(reader.read_byte())
        self.slot = reader.read_7bit_encoded_int()
        self.leashed_type = None
        self.anchor = None
        self.extra_data = b""
        if self.msg_type in (
            LeashedEntityMessageType.FullSync,
            LeashedEntityMessageType.PartialSync,
        ):
            self.leashed_type = reader.read_7bit_encoded_int()
            if self.msg_type == LeashedEntityMessageType.FullSync:
                anchor_x = reader.read_short()
                anchor_y = reader.read_short()
                self.anchor = (anchor_x, anchor_y)
        remaining_len = len(reader.remaining())
        self.extra_data = reader.read_bytes(remaining_len)

    def write(self, writer: Writer) -> None:
        if self.msg_type is None or self.slot is None:
            raise ValueError("msg_type and slot must not be None")
        writer.write_byte(self.msg_type.value)
        writer.write_7bit_encoded_int(self.slot)
        if self.msg_type in (
            LeashedEntityMessageType.FullSync,
            LeashedEntityMessageType.PartialSync,
        ):
            if self.leashed_type is None:
                raise ValueError("leashed_type must not be None for FullSync/PartialSync")
            writer.write_7bit_encoded_int(self.leashed_type)
            if self.msg_type == LeashedEntityMessageType.FullSync:
                if self.anchor is None:
                    raise ValueError("anchor must not be None for FullSync")
                writer.write_short(self.anchor[0])
                writer.write_short(self.anchor[1])
        if self.extra_data is None:
            raise ValueError("extra_data must not be None")
        writer.write_bytes(self.extra_data)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class Emoji(SyncPacket):
    id = MessageID.Emoji

    def __init__(self, player_id: int = 0, emoticon: int = 0):
        self.player_id = player_id
        self.emoticon = emoticon

    def read(self, reader: Reader):
        self.player_id = reader.read_byte()
        self.emoticon = reader.read_byte()

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_byte(self.emoticon)



from dataclasses import dataclass
from enum import IntEnum

from terrex.net.streamer import Reader, Writer

from .net_module import NetServerModule


class BannersMessageType(IntEnum):
    FullState = 0
    KillCountUpdate = 1
    ClaimCountUpdate = 2
    ClaimRequest = 3
    ClaimResponse = 4


@dataclass()
class NetBannersModule(NetServerModule):
    id: int = 10
    msg_type: BannersMessageType | None = None
    banner_id: int | None = None
    kill_count: int | None = None
    claimable_count: int | None = None
    amount: int | None = None
    granted: bool | None = None
    kill_counts: list[int] | None = None
    claimable_banners: list[int] | None = None

    @classmethod
    def create(
        cls,
        msg_type: BannersMessageType,
        banner_id: int | None = None,
        kill_count: int | None = None,
        claimable_count: int | None = None,
        amount: int | None = None,
        granted: bool | None = None,
        kill_counts: list[int] | None = None,
        claimable_banners: list[int] | None = None,
    ) -> 'NetBannersModule':
        obj = cls()
        obj.msg_type = msg_type
        obj.banner_id = banner_id
        obj.kill_count = kill_count
        obj.claimable_count = claimable_count
        obj.amount = amount
        obj.granted = granted
        obj.kill_counts = kill_counts
        obj.claimable_banners = claimable_banners
        return obj

    def read(self, reader: Reader) -> None:
        self.msg_type = BannersMessageType(reader.read_byte())
        self.banner_id = None
        self.kill_count = None
        self.claimable_count = None
        self.amount = None
        self.granted = None
        self.kill_counts = None
        self.claimable_banners = None
        if self.msg_type == BannersMessageType.FullState:
            num = reader.read_short()
            self.kill_counts = [reader.read_int() for _ in range(num)]
            if reader.version >= 289:
                num_claim = reader.read_short()
                self.claimable_banners = [reader.read_ushort() for _ in range(num_claim)]
        elif self.msg_type == BannersMessageType.KillCountUpdate:
            self.banner_id = reader.read_short()
            self.kill_count = reader.read_int()
        elif self.msg_type == BannersMessageType.ClaimCountUpdate:
            self.banner_id = reader.read_short()
            self.claimable_count = reader.read_ushort()
        elif self.msg_type == BannersMessageType.ClaimRequest:
            self.banner_id = reader.read_short()
            self.amount = reader.read_ushort()
        elif self.msg_type == BannersMessageType.ClaimResponse:
            self.banner_id = reader.read_short()
            self.amount = reader.read_ushort()
            self.granted = reader.read_bool()

    def write(self, writer: Writer) -> None:
        if self.msg_type is None:
            raise ValueError("msg_type must not be None")
        writer.write_byte(self.msg_type.value)
        if self.msg_type == BannersMessageType.FullState:
            if self.kill_counts is not None:
                writer.write_short(len(self.kill_counts))
                for kc in self.kill_counts:
                    writer.write_int(kc)
            if self.claimable_banners is not None:
                writer.write_short(len(self.claimable_banners))
                for cb in self.claimable_banners:
                    writer.write_ushort(cb)
        elif self.msg_type == BannersMessageType.KillCountUpdate:
            if self.banner_id is None or self.kill_count is None:
                raise ValueError("banner_id and kill_count must not be None for KillCountUpdate")
            writer.write_short(self.banner_id)
            writer.write_int(self.kill_count)
        elif self.msg_type == BannersMessageType.ClaimCountUpdate:
            if self.banner_id is None or self.claimable_count is None:
                raise ValueError(
                    "banner_id and claimable_count must not be None for ClaimCountUpdate"
                )
            writer.write_short(self.banner_id)
            writer.write_ushort(self.claimable_count)
        elif self.msg_type == BannersMessageType.ClaimRequest:
            if self.banner_id is None or self.amount is None:
                raise ValueError("banner_id and amount must not be None for ClaimRequest")
            writer.write_short(self.banner_id)
            writer.write_ushort(self.amount)
        elif self.msg_type == BannersMessageType.ClaimResponse:
            if self.banner_id is None or self.amount is None or self.granted is None:
                raise ValueError("banner_id, amount, granted must not be None for ClaimResponse")
            writer.write_short(self.banner_id)
            writer.write_ushort(self.amount)
            writer.write_bool(self.granted)



from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.packet.base import ServerPacket


class UpdateGoodEvil(ServerPacket):
    id = MessageID.UpdateGoodEvil

    def __init__(self, good: int = 0, evil: int = 0, crimson: int = 0):
        self.good = good
        self.evil = evil
        self.crimson = crimson

    def read(self, reader: Reader):
        self.good = reader.read_byte()
        self.evil = reader.read_byte()
        self.crimson = reader.read_byte()



from enum import IntEnum

from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.net.structure.vec2 import Vec2
from terrex.packet.base import ServerPacket


class SoundMode(IntEnum):
    STYLE = 1
    VOLUME_SCALE = 2
    PITCH_OFFSET = 3


class PlayLegacySound(ServerPacket):
    id = MessageID.PlayLegacySound

    def __init__(
        self, pos: Vec2 | None = None, sound_id: int = 0, mode: SoundMode = SoundMode.STYLE
    ):
        self.pos = pos or Vec2(0.0, 0.0)
        self.sound_id = sound_id
        self.mode = mode

    def read(self, reader: Reader):
        self.pos = Vec2.read(reader)
        self.sound_id = reader.read_ushort()
        self.mode = SoundMode(reader.read_byte())



import asyncio
import concurrent.futures
import inspect
from collections.abc import Awaitable, Callable
from functools import partial
from typing import Any

from terrex.event.context import EventContext

from .filter.base import EventFilter, StopPropagation
from .types import BaseEvent


class Dispatcher:
    def __init__(self, terrex):
        from terrex.terrex import Terrex

        if not isinstance(terrex, Terrex):
            raise TypeError("terrex must be a Terrex instance")

        self._terrex = terrex
        self._handlers: dict[
            type[BaseEvent],
            list[
                tuple[
                    int, EventFilter[Any], Callable[[Any], Awaitable[None]] | Callable[[Any], None]
                ]
            ],
        ] = {}
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=20)

    def register(
        self,
        filter: EventFilter,
        callback: Callable[[Any], Awaitable[None]] | Callable[[Any], None],
        priority: int = 0,
    ) -> None:
        key = filter._event_type
        self._handlers.setdefault(key, []).append((priority, filter, callback))
        self._handlers[key].sort(key=lambda x: x[0], reverse=True)

    async def dispatch(self, event: BaseEvent) -> None:
        if not self._handlers:
            return

        loop = asyncio.get_running_loop()
        ctx = EventContext(self._terrex, event)

        for _, filter, callback in self._handlers.get(type(event), []):
            matched_event = filter.matches(ctx)
            if matched_event is None:
                continue

            try:
                sig = inspect.signature(callback)
                params = sig.parameters
                args = [matched_event] if len(params) == 1 else []

                if inspect.iscoroutinefunction(callback):
                    await callback(*args)
                else:
                    await loop.run_in_executor(self._executor, partial(callback, *args))
            except StopPropagation:
                break

    def shutdown(self):
        self._executor.shutdown(wait=True)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class AddPlayerBuffPvP(SyncPacket):
    id = MessageID.AddPlayerBuffPvP

    def __init__(self, player_id: int = 0, buff: int = 0, time: int = 0):
        self.player_id = player_id
        self.buff = buff
        self.time = time

    def read(self, reader: Reader):
        self.player_id = reader.read_byte()
        self.buff = reader.read_ushort()
        self.time = reader.read_int()

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_ushort(self.buff)
        writer.write_int(self.time)



from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.packet.base import ServerPacket


class RemoveRevengeMarker(ServerPacket):
    id = MessageID.RemoveRevengeMarker

    def __init__(self, unique_id: int = 0):
        self.unique_id = unique_id

    def read(self, reader: Reader) -> None:
        self.unique_id = reader.read_int()



from terrex.id import MessageID
from terrex.net.bits_byte import BitsByte
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class SyncPlayerZone(SyncPacket):
    id = MessageID.SyncPlayerZone

    def __init__(
        self,
        player_id: int = 0,
        zone1: BitsByte | None = None,
        zone2: BitsByte | None = None,
        zone3: BitsByte | None = None,
        zone4: BitsByte | None = None,
        zone5: BitsByte | None = None,
        town_npc_count: int = 0,
    ):
        self.player_id = player_id
        self.zone1 = zone1 or BitsByte()
        self.zone2 = zone2 or BitsByte()
        self.zone3 = zone3 or BitsByte()
        self.zone4 = zone4 or BitsByte()
        self.zone5 = zone5 or BitsByte()
        self.town_npc_count = town_npc_count

    def read(self, reader: Reader):
        self.player_id = reader.read_byte()
        self.zone1 = BitsByte(reader.read_byte())
        self.zone2 = BitsByte(reader.read_byte())
        self.zone3 = BitsByte(reader.read_byte())
        self.zone4 = BitsByte(reader.read_byte())
        self.zone5 = BitsByte(reader.read_byte())
        self.town_npc_count = reader.read_byte()

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_byte(int(self.zone1))
        writer.write_byte(int(self.zone2))
        writer.write_byte(int(self.zone3))
        writer.write_byte(int(self.zone4))
        writer.write_byte(int(self.zone5))
        writer.write_byte(self.town_npc_count)

    async def handle(self, world, player, evman):
        player.zone.update(
            zone1=self.zone1, zone2=self.zone2, zone3=self.zone3, zone4=self.zone4, zone5=self.zone5
        )



from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.packet.base import ServerPacket


class SyncEmoteBubble(ServerPacket):
    id = MessageID.SyncEmoteBubble

    def __init__(
        self,
        emote_id: int = 0,
        anchor_type: int = 255,
        player_id: int = 0,
        emote_lifetime: int = 0,
        emote: int = 0,
        emote_metadata: int = 0,
    ):
        self.emote_id = emote_id
        self.anchor_type = anchor_type
        self.player_id = player_id
        self.emote_lifetime = emote_lifetime
        self.emote = emote
        self.emote_metadata = emote_metadata

    def read(self, reader: Reader) -> None:
        self.emote_id = reader.read_int()
        self.anchor_type = reader.read_byte()
        if self.anchor_type != 255:
            self.player_id = reader.read_ushort()
            self.emote_lifetime = reader.read_ushort()
            self.emote = reader.read_byte()
            if self.emote < 0:
                self.emote_metadata = reader.read_short()



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ClientPacket


class RequestChestOpen(ClientPacket):
    id = MessageID.RequestChestOpen

    def __init__(self, tile_x: int = 0, tile_y: int = 0):
        self.tile_x = tile_x
        self.tile_y = tile_y

    def write(self, writer: Writer):
        writer.write_short(self.tile_x)
        writer.write_short(self.tile_y)

    def read(self, reader: Reader) -> None:
        self.tile_x = reader.read_short()
        self.tile_y = reader.read_short()



from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.packet.base import ServerPacket


class DeadPlayer(ServerPacket):
    id = MessageID.DeadPlayer

    def __init__(self, player_id: int = 0):
        self.player_id = player_id

    def read(self, reader: Reader):
        self.player_id = reader.read_byte()



"""
Shape module for world generation in Terrex.
"""

from .base import GenAction, GenShape, Point, RectangleArea
from .circle import Circle
from .half_circle import HalfCircle
from .mound import Mound
from .rectangle import Rectangle
from .slime import Slime
from .tail import Tail

__all__ = [
    'Point',
    'RectangleArea',
    'GenAction',
    'GenShape',
    'Rectangle',
    'Circle',
    'HalfCircle',
    'Mound',
    'Slime',
    'Tail',
]



"""
WaterStyleID - autogenerated from decompiled Terraria .cs files
Terraria namespace: Terraria.ID
Terraria version: v1.4.5.5
Generator: TerrariaPyGen v1.1.1 (Maksim Stashkevich)
Generated on: 2026-02-15 09:24
"""

from enum import IntEnum, auto


class WaterStyleID(IntEnum):
    Purity = 0
    Lava = 1
    Corrupt = 2
    Jungle = 3
    Hallow = 4
    Snow = 5
    Desert = 6
    Underground = 7
    Cavern = 8
    Bloodmoon = 9
    Crimson = 10
    Honey = 11
    UndergroundDesert = 12




"""
Ported from Terraria source: Terraria.WorldBuilding.Shapes.Circle
"""

from math import sqrt

import numpy as np

from .base import GenAction, GenActionBulk, GenShape, Point


class Circle(GenShape):
    """Elliptical (oval) shape with independent horizontal and vertical radius."""

    def __init__(
        self, horizontal_radius: int, vertical_radius: int = 0, quit_on_fail: bool = False
    ):
        super().__init__(quit_on_fail)
        if vertical_radius is None:
            vertical_radius = horizontal_radius
        self._horizontalRadius = horizontal_radius
        self._verticalRadius = vertical_radius

    def SetRadius(self, radius: int):
        """Set both horizontal and vertical radius to the given value."""
        self._horizontalRadius = radius
        self._verticalRadius = radius

    def perform(self, origin: Point, action: GenAction) -> bool:
        num = (self._horizontalRadius + 1) ** 2

        # Fast bulk path for GenActionBulk
        if isinstance(action, GenActionBulk):
            xs_all = []
            ys_all = []
            for i in range(origin.y - self._verticalRadius, origin.y + self._verticalRadius + 1):
                y = self._horizontalRadius / self._verticalRadius * (i - origin.y)
                v = num - y * y
                if v < 0:
                    continue
                num1 = min(self._horizontalRadius, int(sqrt(v)))
                xs = np.arange(origin.x - num1, origin.x + num1 + 1)
                ys = np.full_like(xs, i)
                xs_all.append(xs)
                ys_all.append(ys)
            if not xs_all:
                return True
            return action.apply_bulk(np.concatenate(xs_all), np.concatenate(ys_all))

        # Exact C# path (pixel-perfect)
        for i in range(origin.y - self._verticalRadius, origin.y + self._verticalRadius + 1):
            y = self._horizontalRadius / self._verticalRadius * (i - origin.y)
            v = num - y * y
            if v < 0:
                continue
            num1 = min(self._horizontalRadius, int(sqrt(v)))
            for j in range(origin.x - num1, origin.x + num1 + 1):
                if not self.unit_apply(action, origin, j, i) and self._quitOnFail:
                    return False
        return True



import inspect
from abc import ABC, abstractmethod
from typing import Any

from terrex.event.manager import EventManager
from terrex.net.enum.mode import NetMode
from terrex.net.streamer import Reader, Writer
from terrex.player.player import Player
from terrex.util.stringify import stringify_value
from terrex.world.world import World

packet_registry: dict[int, type["Packet"]] = {}


class Packet(ABC):
    id: int

    def read(self, reader: Reader) -> None:
        raise NotImplementedError("Method read must be overridden")

    def write(self, writer: Writer) -> None:
        raise NotImplementedError("Method write must be overridden")

    async def handle(self, world: World, player: Player, evman: EventManager) -> None:
        """Optional method to handle the packet"""
        raise NotImplementedError("Method handle can be overridden if needed")

    def __init_subclass__(cls):
        super().__init_subclass__()

        if inspect.isabstract(cls):
            return

        if not isinstance(getattr(cls, "id", None), int):
            raise TypeError(f"{cls.__name__} must define class attribute 'id'")

        if not callable(getattr(cls, "read", None)):
            raise TypeError(f"{cls.__name__} must implement read()")

        if cls.id in packet_registry:
            raise ValueError(f"Packet id {cls.id} already registered for {packet_registry[cls.id]}")

        packet_registry[cls.id] = cls

    def __to_log_dict(self) -> dict[str, Any]:
        return {name: stringify_value(value) for name, value in vars(self).items()}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__to_log_dict()})"


class ServerPacket(Packet, ABC):
    """Server -> Client packets. Client only reads (read)."""

    _net_mode: NetMode = NetMode.SERVER

    @abstractmethod
    def read(self, reader: Reader) -> None:
        pass

    def write(self, writer: Writer) -> None:
        raise NotImplementedError(
            f"Client does not send {self.__class__.__name__} (server-only packet)"
        )


class ClientPacket(Packet, ABC):
    """Client -> Server packets. Client only writes (write)."""

    _net_mode: NetMode = NetMode.CLIENT

    @abstractmethod
    def write(self, writer: Writer) -> None:
        pass

    def read(self, reader: Reader) -> None:
        raise NotImplementedError(
            f"Client does not read {self.__class__.__name__} (server-bound packet only)"
        )


class SyncPacket(Packet, ABC):
    """Server <-> Client (Sync) packets. Both methods required."""

    _net_mode: NetMode = NetMode.SYNC

    @abstractmethod
    def read(self, reader: Reader) -> None:
        pass

    @abstractmethod
    def write(self, writer: Writer) -> None:
        pass



from dataclasses import dataclass, field
from typing import Any

from terrex.net.streamer import Reader
from terrex.net.tile_stack import TileStack
from terrex.world.world import World


@dataclass
class WorldSection:
    x_start: int = -1
    y_start: int = -1
    width: int = 200
    height: int = 150
    tiles: TileStack = field(default_factory=lambda: TileStack())
    chests: dict[int, Any] = field(default_factory=lambda: {})
    signs: dict[int, Any] = field(default_factory=lambda: {})
    tile_entities: list = field(default_factory=lambda: [])

    def read(self, reader: Reader, world: World) -> None:
        from terrex.entity.tile_entity import read_tile_entity
        from terrex.net.structure.chest import Chest
        from terrex.net.structure.sign import Sign
        from terrex.net.structure.tile import Tile

        tile: Tile | None = None
        rle: int = 0
        for y in range(self.y_start, self.y_start + self.height):
            for x in range(self.x_start, self.x_start + self.width):
                if rle == 0:
                    tile, rle = Tile.read(reader, world, x, y)
                    self.tiles.set(x, y, tile)
                else:
                    rle -= 1
                    copied_tile = world.tiles.get(x, y)
                    if copied_tile is None:
                        copied_tile = Tile()
                    copied_tile.copy_from(tile)
                    world.tiles.set(x, y, copied_tile)
                    self.tiles.set(x, y, copied_tile)

        if rle != 0:
            raise ValueError("WARNING: RLE not exhausted:", rle)

        n_chests = reader.read_short()
        for _ in range(n_chests):
            chest = Chest.read(reader)
            if 0 <= chest.index < 8000:
                world.chests[chest.index] = chest
                self.chests[chest.index] = chest

        n_signs = reader.read_short()
        for _ in range(n_signs):
            sign = Sign.read(reader)
            if 0 <= sign.index < 32000:
                world.signs[sign.index] = sign
                self.signs[sign.index] = sign

        n_entities = reader.read_short()
        for _ in range(n_entities):
            tile_entity = read_tile_entity(reader)
            world.tile_entities.append(tile_entity)
            self.tile_entities.append(tile_entity)

    def __repr__(self) -> str:
        return (
            "WorldSection("
            + ", ".join(
                [
                    f"x_start={self.x_start}",
                    f"y_start={self.y_start}",
                    f"width={self.width}",
                    f"height={self.height}",
                    f"tiles_size={len(self.tiles)}",
                    f"chests_size={len(self.chests)}",
                    f"signs_size={len(self.signs)}",
                    f"tile_entities_size={len(self.tile_entities)})",
                ]
            )
            + ")"
        )



from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.packet.base import ServerPacket


class WiredCannonShot(ServerPacket):
    id = MessageID.WiredCannonShot

    def __init__(
        self,
        damage: int = 0,
        knockback: float = 0.0,
        x: int = 0,
        y: int = 0,
        angle: int = 0,
        ammo: int = 0,
        player_id: int = 0,
    ):
        self.damage = damage
        self.knockback = knockback
        self.x = x
        self.y = y
        self.angle = angle
        self.ammo = ammo
        self.player_id = player_id

    def read(self, reader: Reader) -> None:
        self.damage = reader.read_short()
        self.knockback = reader.read_float()
        self.x = reader.read_short()
        self.y = reader.read_short()
        self.angle = reader.read_short()
        self.ammo = reader.read_short()
        self.player_id = reader.read_byte()



from terrex.id import MessageID
from terrex.net.streamer import Writer
from terrex.packet.base import ClientPacket


class FoodPlatterTryPlacing(ClientPacket):
    id = MessageID.FoodPlatterTryPlacing

    def __init__(self, x: int = 0, y: int = 0, item_id: int = 0, prefix: int = 0, stack: int = 0):
        self.x = x
        self.y = y
        self.item_id = item_id
        self.prefix = prefix
        self.stack = stack

    def write(self, writer: Writer):
        writer.write_short(self.x)
        writer.write_short(self.y)
        writer.write_short(self.item_id)
        writer.write_byte(self.prefix)
        writer.write_short(self.stack)



from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.packet.base import ServerPacket


class TemporaryAnimation(ServerPacket):
    id = MessageID.TemporaryAnimation

    def __init__(self, animation_type: int = 0, tile_type: int = 0, x: int = 0, y: int = 0):
        self.animation_type = animation_type
        self.tile_type = tile_type
        self.x = x
        self.y = y

    def read(self, reader: Reader):
        self.animation_type = reader.read_short()
        self.tile_type = reader.read_ushort()
        self.x = reader.read_short()
        self.y = reader.read_short()



from terrex.id import MessageID
from terrex.net.streamer import Writer
from terrex.packet.base import ClientPacket


class MurderSomeoneElsesPortal(ClientPacket):
    id = MessageID.MurderSomeoneElsesPortal

    def __init__(self, projectile_owner: int = 0, projectile_ai: int = 0):
        self.projectile_owner = projectile_owner
        self.projectile_ai = projectile_ai

    def write(self, writer: Writer):
        writer.write_ushort(self.projectile_owner)
        writer.write_byte(self.projectile_ai)



from enum import IntEnum


class TeleportRequestType(IntEnum):
    TeleportationPotion = 0  # to random pos
    MagicConch = 1
    Shellphone_Ocean = 1
    DemonConch = 2
    Shellphone_Underworld = 2
    Shellphone_Spawn = 3
    PlayerNoSpaceTeleport = 4  # after dismount



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import SyncPacket


class LandGolfBallInCup(SyncPacket):
    id = MessageID.LandGolfBallInCup

    def __init__(
        self, player_id: int = 0, x: int = 0, y: int = 0, number_of_hits: int = 0, proj_id: int = 0
    ):
        self.player_id = player_id
        self.x = x
        self.y = y
        self.number_of_hits = number_of_hits
        self.proj_id = proj_id

    def read(self, reader: Reader):
        self.player_id = reader.read_byte()
        self.x = reader.read_short()
        self.y = reader.read_short()
        self.number_of_hits = reader.read_ushort()
        self.proj_id = reader.read_ushort()

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_short(self.x)
        writer.write_short(self.y)
        writer.write_ushort(self.number_of_hits)
        writer.write_ushort(self.proj_id)



from terrex.id import MessageID
from terrex.net.streamer import Reader
from terrex.packet.base import ServerPacket


class MoonlordHorror(ServerPacket):
    id = MessageID.MoonlordHorror

    def __init__(self, countdown: int = 0):
        self.countdown = countdown

    def read(self, reader: Reader):
        self.countdown = reader.read_int()



from enum import IntEnum

from typing_extensions import Self

from terrex.net.streamer import Reader, Writer


class NetworkTextMode(IntEnum):
    LITERAL = 0
    FORMATTABLE = 1
    LOCALIZATION_KEY = 2

    @classmethod
    def read(cls, reader: Reader) -> 'NetworkTextMode':
        return cls(reader.read_byte())

    def write(self, writer: Writer) -> None:
        writer.write_byte(self.value)


class NetworkText:
    def __init__(
        self,
        mode: NetworkTextMode = NetworkTextMode.LITERAL,
        text: str = "",
        substitutions: list[Self] | None = None,
    ):
        self.mode = mode
        self.text = text
        self.substitutions = substitutions or []

    @classmethod
    def read(cls, reader: Reader) -> 'NetworkText':
        mode = NetworkTextMode.read(reader)
        text = reader.read_dotnet_string()
        substitutions = []
        if mode != NetworkTextMode.LITERAL:
            count = reader.read_byte()
            for _ in range(count):
                substitutions.append(NetworkText.read(reader))
        return cls(mode, text, substitutions)

    def write(self, writer: Writer) -> None:
        self.mode.write(writer)
        writer.write_dotnet_string(self.text)
        if self.mode != NetworkTextMode.LITERAL:
            writer.write_byte(len(self.substitutions))
            for sub in self.substitutions:
                sub.write(writer)



from terrex.id import MessageID
from terrex.net.streamer import Writer
from terrex.packet.base import ClientPacket


class QuestsCountSync(ClientPacket):
    id = MessageID.QuestsCountSync

    def __init__(self, player_id: int = 0, angler_quests_completed: int = 0, golfer_score: int = 0):
        self.player_id = player_id
        self.angler_quests_completed = angler_quests_completed
        self.golfer_score = golfer_score

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_int(self.angler_quests_completed)
        writer.write_int(self.golfer_score)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ClientPacket


class PlayerSpawn(ClientPacket):
    id = MessageID.PlayerSpawn

    def __init__(
        self,
        player_id: int = 0,
        spawn_x: int = -1,
        spawn_y: int = -1,
        respawn_time_remaining: int = 0,
        number_of_deaths_pve: int = 0,
        number_of_deaths_pvp: int = 0,
        team_id: int = 0,
        player_spawn_context: int = 0,
    ):
        self.player_id = player_id
        self.spawn_x = spawn_x
        self.spawn_y = spawn_y
        self.respawn_time_remaining = respawn_time_remaining
        self.number_of_deaths_pve = number_of_deaths_pve
        self.number_of_deaths_pvp = number_of_deaths_pvp
        self.team_id = team_id
        self.player_spawn_context = player_spawn_context

    def write(self, writer: Writer):
        writer.write_byte(self.player_id)
        writer.write_short(self.spawn_x)
        writer.write_short(self.spawn_y)
        writer.write_int(self.respawn_time_remaining)
        writer.write_short(self.number_of_deaths_pve)
        writer.write_short(self.number_of_deaths_pvp)
        writer.write_byte(self.team_id)
        writer.write_byte(self.player_spawn_context)

    def read(self, reader: Reader) -> None:
        self.player_id = reader.read_byte()
        self.spawn_x = reader.read_short()
        self.spawn_y = reader.read_short()
        self.respawn_time_remaining = reader.read_int()
        self.number_of_deaths_pve = reader.read_short()
        self.number_of_deaths_pvp = reader.read_short()
        self.team_id = reader.read_byte()
        self.player_spawn_context = reader.read_byte()



from enum import IntEnum

from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ClientPacket


class TileEntityType(IntEnum):
    TRAINING_DUMMY = 0
    ITEM_FRAME = 1
    LOGIC_SENSOR = 2

    @classmethod
    def read(cls, reader: Reader) -> 'TileEntityType':
        return cls(reader.read_byte())

    def write(self, writer: Writer):
        writer.write_byte(self.value)


class TileEntityPlacement(ClientPacket):
    id = MessageID.TileEntityPlacement

    def __init__(
        self,
        x: int = 0,
        y: int = 0,
        tile_entity_type: TileEntityType = TileEntityType.TRAINING_DUMMY,
    ):
        self.x = x
        self.y = y
        self.tile_entity_type = tile_entity_type

    def write(self, writer: Writer):
        writer.write_short(self.x)
        writer.write_short(self.y)
        self.tile_entity_type.write(writer)

    def read(self, reader: Reader):
        self.x = reader.read_short()
        self.y = reader.read_short()
        self.tile_entity_type = TileEntityType.read(reader)



from terrex.id import MessageID
from terrex.net.streamer import Reader, Writer
from terrex.packet.base import ClientPacket


class GemLockToggle(ClientPacket):
    id = MessageID.GemLockToggle

    def __init__(self, x: int = 0, y: int = 0, on: bool = False):
        self.x = x
        self.y = y
        self.on = on

    def write(self, writer: Writer):
        writer.write_short(self.x)
        writer.write_short(self.y)
        writer.write_bool(self.on)

    def read(self, reader: Reader):
        self.x = reader.read_short()
        self.y = reader.read_short()
        self.on = reader.read_bool()



"""
LiquidID - autogenerated from decompiled Terraria .cs files
Terraria namespace: Terraria.ID
Terraria version: v1.4.5.5
Generator: TerrariaPyGen v1.1.1 (Maksim Stashkevich)
Generated on: 2026-02-15 09:24
"""

from enum import IntEnum, auto


class LiquidID(IntEnum):
    Water = 0
    Lava = 1
    Honey = 2
    Shimmer = 3

